;last update: 2014-04-30 by Andrea La Camera
;+
;-

pro patch_deconv, par, SAVE_OUT=save_out ;, obj
  
  COMMON image, img, header, N, M, a_dim, Nori, Mori
  COMMON background, bg                   ;bg must be N x M, as the input image "img" is.
  COMMON output, out, res_array, residual ; reconstructed object with original dimension Nori x Mori
                                ; res_array is the residual statistics (one for each sub-domain)
                                ; residual is the final map of residuals  
  on_error, 2
;save several information in header
  fxaddpar, header, 'IMG     ',par.img_filename,"Image: blurred image filename"
  if par.bg_choice EQ 0 then begin
     fxaddpar, header, 'SKY BG  ',par.bg, "Image: sky background [ADU/pixel]"
  endif else begin
     fxaddpar, header, 'SKY BG  ',par.bg_file, "Image: sky background from file"
  endelse
  fxaddpar, header, 'RON     ',par.sigma_ron, "Image: RON value [e-/pixel]"
  fxaddpar, header, 'GAIN    ',par.gain, "Image: GAIN value [e-/ADU]"
  fxaddpar, header, 'N_COLS  ',par.cols, "Sub-domain: number of columns"  
  fxaddpar, header, 'N_ROWS  ',par.rows, "Sub-domain: number of rows"  
  fxaddpar, header, 'OVERLAP ',par.overlap, "Sub-domain: overlap"
  fxaddpar, header, 'PSF_TYPE',par.psf_type, "PSF: type (0=cube, 1=single files)"
  if par.psf_type EQ 0 then begin
     fxaddpar, header, 'PSF_CUBE',(*par.psf_filename)[0], "PSF: filename"
  endif else begin
     for k=0,par.cols*par.rows-1 do begin
        fxaddpar, header, 'PSF#'+strtrim(k,1),(*par.psf_filename)[k], "PSF: filename"
     endfor
  endelse
  fxaddpar, header, 'EE_THRES',par.threshold, "PSF: Enclosed Energy threshold"
  fxaddpar, header, 'DECONV  ',par.method, "Deconvolution: method (0=RL,1=SGP)"
  fxaddpar, header, 'STOP_CRI',par.stop_crit, "Deconvolution: stopping rule"
  fxaddpar, header, 'TOL     ',par.tol, "Deconvolution: Tolerance (if requested)"
  
;evaluate the median of the background  
  bg_med=median(bg)
  
;zero-padding of image (in order to consider overlap)
  NN=N+2*par.overlap
  MM=M+2*par.overlap
  img2=zero_padding(img,NN,MM)   
;obj2=zero_padding(obj,NN,MM)
;zero-padding of the background array  
  bg2=zero_padding(bg,NN,MM)
  
;add the median bg value to the zero-padded img
  index=where(img2 LE 0, count)
  if count GT 0 then img2[index]=bg_med 
  
;add the median bg value to the zero-padded bg
  index=where(bg2 LE 0, count)
  if count GT 0 then bg2[index]=bg_med 
  
;add the Read-Out Noise variance to img2 and bg2 (Gaussian+Poisson noise correction)
  img2+=((par.sigma_ron/par.gain)^2.)
  bg2+=((par.sigma_ron/par.gain)^2.)
  bg_med+=((par.sigma_ron/par.gain)^2.)
  
  
;output image (the reconstructed object), residual and other stuff
  out=make_array(N,M,/float)
  residual=make_array(N,M,/float)
  res_array=make_array(par.cols*par.rows,5)
  
;dimension of each sub-domain (+ overlap)
  NS=par.domx+2*par.overlap
  MS=par.domy+2*par.overlap
  
;read psfs
  if par.psf_type EQ 0 then psf=readfits((*par.psf_filename)[0],/SILENT)

;read MASK for PROJECTION (in SGP)
  if (par.mask_choice) then begin
     pr_mask=readfits(par.mask_filename, /SILENT)
     pr_mask=zero_padding(pr_mask,NN,MM,value=1.)
  endif else begin
     pr_mask=make_array(NN,MM,value=1.)
  endelse
 
  tt=systime(1)
;-------> RUN!
  for k=0,par.cols*par.rows-1 do begin
     t0=systime(1)
                                ;cycle on sub-domains
                                ;selection of current sub-domain 
     x0=(k mod par.cols)*par.domx
     y0=(k / par.cols)*par.domy
     x1=x0+NS-1
     y1=y0+MS-1
     g=img2[x0:x1,y0:y1]
     b=bg2[x0:x1,y0:y1]
     sub_pr_mask=pr_mask[x0:x1,y0:y1]
                                ;o=obj2[x0:x1,y0:y1]
                                ;flux evaluation
     flux=total(g-b, /DOUBLE)
                                ;selection of current PSF
     if par.psf_type EQ 1 then begin
        psf_curr=readfits((*par.psf_filename)[k],/SILENT)
     endif else psf_curr=psf[*,*,k]
                                ;(re-)normalization of the psf
     psf_curr=psf_curr/total(psf_curr,/double)
     NP=(size(psf_curr))[1]
     MP=(size(psf_curr))[2]
                                ;compute Delta (the Enclosed Energy crosses the threshold at distance = Delta) 
;delta=patch_psf_ee(psf_curr, par.threshold)
                                ;print, "DELTA = ", delta
;NS1=NS+2*delta
;MS1=MS+2*delta
     case par.method of
        0: begin
                                ;RICHARDSON-LUCY METHOD + BOUNDARY
                                ;resizing
;dim=max([NS1,MS1,NP,MP])
           dim=max([NS,MS,NP,MP])
           dim= dim + dim mod 2 ; I want an even number !
           if dim LT 2l^ceil(alog(dim)/alog(2.)) then dim = 2l^ceil(alog(dim)/alog(2.))
           psf_curr=zero_padding(psf_curr, dim, dim) 
           g=zero_padding(g,dim,dim)
           b=zero_padding(b,dim,dim,value=bg_med)
           sub_pr_mask=zero_padding(sub_pr_mask,dim,dim,value=1.)
           print, "Working dimensions = ", dim
                                ;FFT of the PSF
           tf_psf=FFT(shift(psf_curr,dim/2,dim/2),-1,/double)*dim*dim
                                ;computation of the initial arrays
           mask=make_array(x1-x0+1,y1-y0+1, value=1.)
           mask=zero_padding(mask,dim,dim)
           alpha=FLOAT(FFT(conj(tf_psf)*FFT(mask,-1,/double),+1,/double))
           w=1./alpha
           index=where(alpha LT 1e-2, count)
           if count GT 0 then w[index]=0.
                                ;initial rec --- with exact flux in the reconstruction region
           rec=make_array(dim,dim,value=(flux/total(alpha,/DOU)))
           if count GT 0 then rec[index]=0.
           rec=rec*sub_pr_mask
                                ;run RL iterations  
           iter=0
           print, 'DECONV RL --> nb of iterations = ', (*par.tab_iter)[k]
           patch_deconv_rl, g,tf_psf,b,w,(*par.tab_iter)[k],par.stop_crit,rec, iter, res, res_stat
                                ;save rec'd object into correct
                                ;position 
           out[x0:x0+par.domx-1,y0:y0+par.domy-1]=rec[(dim-par.domx)/2:(dim+par.domx)/2-1,$
                                                      (dim-par.domy)/2:(dim+par.domy)/2-1]
                                ; save residual into array 
           residual[x0:x0+par.domx-1,y0:y0+par.domy-1]=res[(dim-par.domx)/2:(dim+par.domx)/2-1,$
                                                           (dim-par.domy)/2:(dim+par.domy)/2-1]
           res_array[k,*]=res_stat
           print, " --> sub-domain #"+strtrim(k,1)+" done!  (elapsed time:"+$
                  strtrim(systime(1)-t0,1)+" sec)  (it = "+strtrim(iter,1)+")"
        end
        1: begin
                                ; SGP METHOD
                                ; if g is very small (i.e. less than 2^(N-1)) while the PSF is less than
                                ; or equal to 2^N) . For istance: g=200x200 and PSF = 400x400 --> 
                                ; g=256x256 and PSF = 512x512. Next, SGP will extend g to the actual size of
                                ; the PSF (512).  
;if (max([NS1,MS1]) LE 2^floor(alog(max([NP,MP]))/alog(2))) then begin
           if (max([NS,MS]) LE 2^floor(alog(max([NP,MP]))/alog(2))) then begin
              g=zero_padding(g,2^floor(alog(max([NP,MP]))/alog(2)),2^floor(alog(max([NP,MP]))/alog(2)),value=bg_med)
              b=zero_padding(b,2^floor(alog(max([NP,MP]))/alog(2)),2^floor(alog(max([NP,MP]))/alog(2)),value=bg_med)
              sub_pr_mask=zero_padding(sub_pr_mask,2^floor(alog(max([NP,MP]))/alog(2)),2^floor(alog(max([NP,MP]))/alog(2)),value=1.)
                                ;o=zero_padding(o,2^floor(alog(max([NP,MP]))/alog(2)), 2^floor(alog(max([NP,MP]))/alog(2)))
;              NS1=(size(g))[1]
;              MS1=(size(g))[2]
              NS=(size(g))[1]
              MS=(size(g))[2]
;              dim=max([NS1,MS1,NP,MP])
              dim=max([NS,MS,NP,MP])
              if dim LT 2l^ceil(alog(dim)/alog(2.)) then dim = 2l^ceil(alog(dim)/alog(2.))
              psf_curr=zero_padding(psf_curr, dim, dim)
                                ;o=zero_padding(o, dim, dim)
           endif else begin     ; in all other cases enlarge g,b,and PSF to the same dim.        
;              dim=max([NS1,MS1,NP,MP])
              dim=max([NS,MS,NP,MP])
              dim= dim + dim mod 2 ; I want an even number !
              psf_curr=zero_padding(psf_curr, dim, dim)
              g=zero_padding(g,dim,dim,value=bg_med)
              b=zero_padding(b,dim,dim,value=bg_med)
              sub_pr_mask=zero_padding(sub_pr_mask,dim,dim,value=1.)
                                ;o=zero_padding(o,dim,dim)
           endelse
           print, "Working dimensions = ", dim
           iter=0
           ;print, 'DECONV SGP --> nb of iterations = ', (*par.tab_iter)[k]
           patch_deconv_sgp, g, psf_curr, b, (*par.tab_iter)[k], $
                                  par.stop_crit, par.tol, rec, iter, sub_pr_mask, res, res_stat
                                ;save rec'd object into correct
                                ;position 
           iter=iter+1
           dim=(size(rec))[1]
           out[x0:x0+par.domx-1,y0:y0+par.domy-1]=rec[(dim-par.domx)/2:(dim+par.domx)/2-1,$
                                                      (dim-par.domy)/2:(dim+par.domy)/2-1]
                                ; save residual into array 
           residual[x0:x0+par.domx-1,y0:y0+par.domy-1]=res[(dim-par.domx)/2:(dim+par.domx)/2-1,$
                                                           (dim-par.domy)/2:(dim+par.domy)/2-1]
           
           res_array[k,*]=res_stat
           print, " --> sub-domain #"+strtrim(k,1)+" done!  (elapsed time:"+$
                  strtrim(systime(1)-t0,1)+" sec)  (it = "+strtrim(iter-1,1)+")"
                                ;print, "err = ", err
                                ;print, "Discr = ", Discr
        end
        else: begin
           message, "Deconvolution method not implemented!"
        endelse
     endcase
;add other info to header
     fxaddpar, header, "DOM#"+strtrim(k,1),iter-1 , "Deconvolution: iterations in "+strtrim(systime(1)-t0,1)+"(sec)" 
  endfor
;last header update
  fxaddpar,header, "DEC_TIME",strtrim(systime(1)-tt,1),"Deconvolution: Total time (sec)" 
  
;crop reconstructed image to input (original) dimensions
  out=out[(N-Nori)/2:(N+Nori)/2-1, (M-Mori)/2:(M+Mori)/2-1]
;crop the residual image to input (original) dimensions
  residual=residual[(N-Nori)/2:(N+Nori)/2-1, (M-Mori)/2:(M+Mori)/2-1]
  
;save output in FITS file if SAVE_OUT keyword is set
  if KEYWORD_SET(save_out) then begin
     writefits, par.out_filename, out, header
     writefits, par.res_filename, residual, header
  endif
;ok! done!
end
